import datetime
import pandas as pd
import numpy as np
import scipy.optimize as sco
from numpy.linalg import inv
import math

class ModelCalulator:

    STATS_DICT = {
        'ret':0,
        'std':1,
        'sharpe':2,
    }

    def __init__(self, risk_free:float=0.03, trading_day_per_year:int=242):
        """
        - ''risk_free'' - (float) risk free rate default 0.03
        - ''trading_day_per_year'' -(int) trading day each year 242 or 252
        """
        self.risk_free = risk_free
        self.trading_day_per_year = trading_day_per_year

    def _calc_statistics(self, weights, *args):    
        returns = args[0]
        weights = np.array(list(weights))
        if len(args) == 1:
            days = (returns.index[-1] - returns.index[0]).days
            _mv = (returns * weights).sum(axis=1)
            last_nav = (_mv + 1).prod()
            port_ret = math.pow(last_nav,365/days)-1
            port_std = _mv.std() * np.sqrt(242)
        else:
            ret_mean = args[1]
            ret_cov = args[2]
            port_ret = np.sum(ret_mean * weights)*self.trading_day_per_year
            port_std = np.sqrt(np.dot(weights.T, np.dot(ret_cov*self.trading_day_per_year,weights)))

        return np.array([port_ret, port_std, (port_ret-self.risk_free)/port_std])

    def _cons_parse(self, returns:pd.DataFrame, asset_weight_cons:list=[], port_stats_cons:list=[], weight_stats_cons:list=[]):
        asset_list = returns.columns.tolist() 
        # 默认权重合为1 不允许做空
        cons = ({'type': 'eq', 'fun': self._total_weight_constraint},
                {'type': 'ineq', 'fun': self._long_only_constraint})
        
        # 单资产风险权重条件
        if len(weight_stats_cons) == 0:
            pass
        else:
            #  但资产风险权重下下 现金权重不超过3%
            cons += ({'type':'ineq','fun':lambda x: -x[-1] - 0.03},)
            risk_list = weight_stats_cons[0]
            low_list = weight_stats_cons[1]
            up_list = weight_stats_cons[2]

            for idx, asset_id in enumerate(asset_list[:-1]):
                print(f'asset {idx} risk weight <= {up_list[idx]}')
                print(f'up_list {up_list} idx {idx} risk_list {risk_list} asset_list {asset_list}')
                # 风险权重不包含现金设置
                ## 风险占比上限 * 总风险 >= 子基金权重 * 子基金风险 
                cons += ({'type':'ineq','fun':lambda x: up_list[idx] * sum(np.array(risk_list)*np.array(x[:-1])) - risk_list[idx] * x[idx]}, )
                ## 风险占比下限
                cons += ({'type':'ineq','fun':lambda x: -low_list[idx] * sum(np.array(risk_list)*np.array(x[:-1])) + risk_list[idx] * x[idx]}, )

        # 组合指标条件
        for port_stats_con_i in port_stats_cons:
            stat_name = port_stats_con_i[0]
            method = port_stats_con_i[1]
            num = port_stats_con_i[2]
            idx = self.STATS_DICT[stat_name]
            print(stat_name, method, num, idx)
            if method == '=':
                cons += ({'type':'eq','fun':lambda x: self._calc_statistics(x,(returns))[idx] - num}, )
            elif method == '>=':
                cons += ({'type':'ineq','fun':lambda x: self._calc_statistics(x,(returns))[idx] - num}, )
            elif method == '<=':
                cons += ({'type':'ineq','fun':lambda x: -self._calc_statistics(x,(returns))[idx] + num}, )

        # 资产权重条件
        for asset_weight_con_i in asset_weight_cons:
            asset_id = asset_weight_con_i[0]
            method = asset_weight_con_i[1]
            num = asset_weight_con_i[2]
            asset_idx = asset_list.index(asset_id)
            print(asset_id, method, num, asset_idx)
            if method == '=':
                cons += ({'type':'eq','fun':lambda x: x[asset_idx] - num},)
            elif method == '>=':
                cons += ({'type':'ineq','fun':lambda x: x[asset_idx] - num},) 
            elif method == '<=':
                cons += ({'type':'ineq','fun':lambda x: -x[asset_idx] + num},) 
        
        return cons

    def _view_parse(self, investor_views:list, returns:pd.DataFrame, market_weights:list):
        def _asset_weight_sum(assets, market_weights, asset_list):
            return sum(market_weights[asset_list.index(asset_id)] for asset_id in assets)
        
        asset_list = returns.columns.tolist()
        views = []
        pick_list = []
        view_confidences = []
        for view_i in investor_views:
            good_assets = view_i[0]
            bad_assets = view_i[1]
            ret = view_i[2]
            confidence = view_i[3]
            views.append(ret)
            view_confidences.append(confidence)
            pick_dict = {}
            good_asset_weights = _asset_weight_sum(good_assets, market_weights, asset_list)
            bad_asset_weights = _asset_weight_sum(bad_assets, market_weights, asset_list)
            for asset_id in good_assets:
                w = market_weights[asset_list.index(asset_id)]
                pick_dict[asset_id] = w / good_asset_weights
            for asset_id in bad_assets:
                w = market_weights[asset_list.index(asset_id)]
                pick_dict[asset_id] = - w/ bad_asset_weights
            pick_list.append(pick_dict)
        return views, pick_list, view_confidences    

    def _max_sharpe(self, weights, *args):   
        return -self._calc_statistics(weights, *args)[2]

    def _min_vol(self, weights, *args):
        return self._calc_statistics(weights, *args)[1]
        
    def _max_ret(self, weights, *args):
        return -self._calc_statistics(weights, *args)[0]

    def _get_opt_weight(self, opts, fund_list):
        return pd.DataFrame(np.reshape(opts.x.round(3), newshape=(1,len(opts.x))), columns=fund_list)

    def _get_port_stats(self, opts, returns):
        return pd.DataFrame(np.reshape(self._calc_statistics(opts.x,(returns)).round(3), newshape=(1,3)),columns=['年化收益','年化波动','夏普率'])

    def _calculate_risk(self, risk_method, returns, fund_nav, day_per_year):
        if risk_method == 'annual_vol':
            risk = returns.std(ddof=1)*np.sqrt(day_per_year)
        elif risk_method == 'mdd':
            risk = 1 - (fund_nav / fund_nav.cummax()).min()
        elif risk_method == 'var':
            risk = returns.quantile(0.05)
        elif risk_method == 'cvar':
            var = returns.quantile(0.05)
            risk = returns[returns < var].mean()
        risk = np.array(risk)
        return risk
        
    def _risk_budget_objective(self, x, *pars):
        risk_method = pars[0]           
        target_risk_weight = pars[1]         
        returns = pars[2] 
        fund_nav = pars[3]    
        day_per_year = pars[4]

        risk = self._calculate_risk_contribution(x, risk_method, returns, fund_nav, day_per_year)
        risk = risk * 100
        s2 = np.sum(target_risk_weight)
        target_risk_weight = target_risk_weight/s2*100
        s = np.square(risk-target_risk_weight.T)
        J = np.sum(np.diagonal(s))
        return J

    def _calculate_risk_contribution(self, x, risk_method, returns, fund_nav, day_per_year):
        risk = self._calculate_risk(risk_method, returns, fund_nav, day_per_year)
        risk = risk * np.array(x)
        risk.resize((risk.shape[0],1))
        risk = np.matrix(risk)
        s1 = np.sum(risk)
        risk = risk/s1
        return risk

    def _calculate_risk_budget_var_corr_method(self, w, V):
        w = np.matrix(w)
        sigma = (w*V*w.T)[0,0]
        return sigma

    def _calculate_risk_budget_var_contribution_corr_method(self, w, V):
        # function that calculates asset contribution to total risk
        w = np.matrix(w)
        sigma = np.sqrt(self._calculate_risk_budget_var_corr_method(w,V))
        # Marginal Risk Contribution
        MRC = V*w.T/sigma
        # Risk Contribution
        RC = np.multiply(MRC,w.T)
        return RC

    def _calculate_risk_budget_objective_corr_method(self, x,pars):
        # calculate portfolio risk
        V = pars[0]# covariance table
        x_t = pars[1] # risk target in percent of portfolio risk
        sig_p =  np.sqrt(self._calculate_risk_budget_var_corr_method(x,V)) # portfolio sigma
        risk_target = np.asmatrix(np.multiply(sig_p,x_t)) #* 100 TODO
        asset_RC = self._calculate_risk_budget_var_contribution_corr_method(x,V) #* 100 TODO
        J = sum(np.square(asset_RC-risk_target.T))[0,0] # sum of squared error
        return J   

    def _total_weight_constraint(self,x):
        return np.sum(x)-1.0

    def _long_only_constraint(self,x):
        return x

    def get_opt_weight(self):
        return self.weights
    
    def get_ports_stats(self):
        return self.port_stats

    def get_asset_risk(self):
        return self.port_asset_risk


class MeanVarianceModel(ModelCalulator):

    def __init__(self,risk_free:float=0.03, trading_day_per_year:int=242): 
        super().__init__(risk_free=risk_free, trading_day_per_year=trading_day_per_year)
        
    def process(self, returns:pd.DataFrame,opt_method:str,asset_weight_cons:list=[],port_stats_cons:list=[], weight_stats_cons:list=[]):
        """
        均值方差模型

        - ''returns'' - (pd.DataFrame) columns name is asset name
        - ''opt_method'' str,  in ['max_sharpe', 'limit_std', 'max_ret']
        - ''asset_weight_cons'' - (list of list) 某资产权重限制 
            每单元由三部分组成： [000001!0, '<=', 0.6]
                        1 资产id （str） 
                        2 大小方式（str） ’=’ ‘>=’ '<='
                        3 权重数  (float)    
        - ''port_stats_cons'' - (list of list) 组合指标限制
            每单元由三部分组成：['std', '<=', 0.2]
                        1 指标：['ret','std','sharpe']
                        2 大小方式（str） ’=’ ‘>=’ '<='
                        3 指标数  (float)                      
        """ 
        cons = self._cons_parse(returns, asset_weight_cons, port_stats_cons, weight_stats_cons)
        # 增加资产cash
        returns.loc[:,'cash'] = 0
        asset_list = returns.columns.tolist()
        noa = len(asset_list)
        bnds = tuple((0,1) for x in range(noa)) # 权重限制在0,1之间 不允许做空
        if opt_method == 'max_sharpe':
            opt_func = self._max_sharpe
        elif opt_method == 'max_ret':
            opt_func = self._max_ret
        elif opt_method == 'min_std':
            opt_func = self._min_vol
        opts = sco.minimize(fun=opt_func, args=(returns), x0=noa*[1./noa,], method='SLSQP', bounds=bnds, constraints=cons)
        self.weights = self._get_opt_weight(opts, asset_list)
        self.port_stats = self._get_port_stats(opts, returns)

class BlackLittermanModel(ModelCalulator):

    def __init__(self,risk_free:float=0.03, trading_day_per_year:int=242): 
        super().__init__(risk_free=risk_free, trading_day_per_year=trading_day_per_year)

    def process(self,   returns:pd.DataFrame,
                        market_weights:list,
                        investor_views:list,
                        tau:float,
                        opt_method:str,
                        risk_aversion:float,
                        omega_method:str,
                        asset_weight_cons:list=[],
                        port_stats_cons:list=[]):
        """
        Black Litterman 模型             
        - ''returns'' - (pd.DataFrame) columns name is asset name
        - ''asset_list'' - (list) asset name
        - ''market_weights'' - (list) weight of asset [36, 8, 56] i.e. sum up to 100
        - ''investor_views'' - (list of list) investor view information list
        ‘’看好的资产‘’ 比 ‘’看衰的资产‘’ 收益高 ‘’观点内的收益‘’， 信心水平 ''信心水平'' 
            每单元由四部分组成：[[],[],float,float]
                        1. 看好的资产列表 list 
                        2. 看衰的资产列表 list 
                                            如果看衰的资产列表非空，表示观点是相对观点 As 好于 Bs
                                            如果看衰的资产列表为空，表示观点是绝对观点 As 好
                        3. 观点内的收益 float 
                        4. 信心水平    float
            #TODO 如果排序 等分排序 
        - ''tau'' - (float) investor view parameter less than 0.05,
        - ''opt_method'' - (str), 'max_sharpe' 'max_ret' 'min_std'
        - ''risk_aversion'' - (float) risk aversion, expected measures of returns with one measure of risk
        - ''omega_method'' - (str) 'prior_variance' or 'user_confidences'
                            prior_variance: investor confidence calculated by market history
                            user_confidences: investor confidence provide by user
        - ''asset_weight_cons'' - (list of list) 某资产权重限制 
            每单元由三部分组成： [000001!0, '<=', 0.6]
                        1 资产id （str） 
                        2 大小方式（str） ’=’ ‘>=’ '<='
                        3 权重数  (float)    
        - ''port_stats_cons'' - (list of list) 组合指标限制
            每单元由三部分组成：['std', '<=', 0.2]
                        1 指标：['ret','std','sharpe']
                        2 大小方式（str） ’=’ ‘>=’ '<='
                        3 指标数  (float)       
        """
        investor_views, pick_list, view_confidences = self._view_parse(investor_views,returns,market_weights)
        cons = self._cons_parse(returns, asset_weight_cons, port_stats_cons)

        # step 0 data prepare
        returns.loc[:,'cash'] = 0
        asset_list = returns.columns.tolist()
        market_capitalised_weights = market_weights + [0]
        covariance = returns.cov()
        asset_names = returns.columns.tolist()
        noa = len(asset_list)
        bnds = tuple((0,1) for x in range(noa)) # 权重限制在0,1之间 不允许做空

        # step 1
        num_assets = len(market_capitalised_weights)
        num_views = len(view_confidences)
        if asset_names is None:
            if covariance is not None and isinstance(covariance, pd.DataFrame):
                asset_names = covariance.columns
            else:
                asset_names = list(map(str, range(num_assets)))

        # step 2 pre process inputs
        investor_views = np.array(np.reshape(investor_views, newshape=(len(investor_views), 1)))
        market_capitalised_weights = np.array(np.reshape(market_capitalised_weights, newshape=(len(market_capitalised_weights), 1)))
        if isinstance(covariance, pd.DataFrame):
            covariance = covariance.values

        # step 3 calculate implied equilbrium returns
        implied_equilibrium_returns = risk_aversion * covariance.dot(market_capitalised_weights)

        # step 4 create the pick matrix (P) from user specified assets involved in the views
        pick_matrix = np.zeros((num_views, num_assets))
        pick_matrix = pd.DataFrame(pick_matrix, columns=asset_names)
        for view_index, pick_dict in enumerate(pick_list):
            assets = list(pick_dict.keys()) 
            values = list(pick_dict.values())
            pick_matrix.loc[view_index, assets] = values
        pick_matrix = pick_matrix.values

        # step 5 build the covariance matrix of errors in investor views (omega)
        if omega_method == 'prior_variance':
            print('without view confidences')
            omega = pick_matrix.dot((tau * covariance).dot(pick_matrix.T))
        else:
            print('include detault view confidences ')
            view_confidences = np.array(np.reshape(view_confidences, (1, len(view_confidences))))
            alpha = (1 - view_confidences) / view_confidences
            omega = alpha * pick_matrix.dot(covariance).dot(pick_matrix.T)
        omega = np.diag(np.diag(omega))
        omega = np.array(np.reshape(omega, (num_views, num_views)))

        # step 6 BL expected returns
        posterior_expected_returns = implied_equilibrium_returns + (tau * covariance).dot(pick_matrix.T).\
            dot(inv(pick_matrix.dot(tau * covariance).dot(pick_matrix.T) + omega).dot(investor_views - pick_matrix.dot(implied_equilibrium_returns)))
        posterior_expected_returns = posterior_expected_returns.reshape(1, -1)

        # step 7 BL covariance
        posterior_covariance = covariance + (tau * covariance) - (tau * covariance).dot(pick_matrix.T).\
                    dot(inv(pick_matrix.dot(tau * covariance).dot(pick_matrix.T) + omega)).dot(pick_matrix).dot(tau * covariance)

        if opt_method == 'max_sharpe':
            opt_func = self._max_sharpe
        elif opt_method == 'max_ret':
            opt_func = self._max_ret
        elif opt_method == 'min_std':
            opt_func = self._min_vol
        opts = sco.minimize(fun=opt_func, args=(returns,posterior_expected_returns,posterior_covariance), x0=noa*[1./noa,], method='SLSQP', bounds=bnds, constraints=cons)
        self.weights = self._get_opt_weight(opts, asset_list)
        self.port_stats = self._get_port_stats(opts, returns)

class RiskParityModel(ModelCalulator):

    def __init__(self,risk_free:float=0.03, trading_day_per_year:int=242): 
        super().__init__(risk_free=risk_free, trading_day_per_year=trading_day_per_year)

    def process(self, risk_method:str, fund_nav:pd.DataFrame, targets:list=[]):
        """
        风险平价模型  基于收益率序列
        - ''risk_method'' - (str) risk measure: annual_vol, mdd, var, cvar
        - ''fund_nav'' - (pd.DataFrame) fund nav, columns is asset name
        - ''targets'' - (list) risk budget, if is None, turn into Risk Parity Model
        """
        returns = fund_nav.pct_change(1).iloc[1:]
        if targets == []:
            targets = returns.shape[1] * [1/returns.shape[1]]
        x_0 = returns.shape[1] * [1/returns.shape[1]]
        cons = ({'type': 'eq', 'fun': self._total_weight_constraint},
        {'type': 'ineq', 'fun': self._long_only_constraint})
        res= sco.minimize(fun=self._risk_budget_objective, x0=x_0, args=(risk_method,targets,returns,fund_nav,self.trading_day_per_year), method='SLSQP',constraints=cons, options={'disp': False})
        weights = list(res.x)
        port_asset_risk = self._calculate_risk_contribution(weights, risk_method, returns, fund_nav, self.trading_day_per_year)
        self.port_asset_risk = list(np.array(port_asset_risk).T[0])
        self.weights = {}
        asset_list = returns.columns.tolist()
        for idx, w in enumerate(weights):
            self.weights[asset_list[idx]] = round(w,4)
        self.port_stats = self._get_port_stats(res, returns)

class RiskBudgetModel(ModelCalulator):

    def __init__(self, risk_free:float=0.03, trading_day_per_year:int=242):
        super().__init__(risk_free=risk_free, trading_day_per_year=trading_day_per_year)

    def process(self, fund_nav:pd.DataFrame, targets:list=[],asset_weight_cons:list=[],port_stats_cons:list=[]):
        """
        风险预算模型  基于协方差矩阵
         - ''fund_nav'' - (pd.DataFrame) fund nav, columns is asset name
        - ''targets'' - (list) risk budget, if is None, turn into Risk Parity Model
        """
        returns = fund_nav.pct_change(1).iloc[1:]
        cons = self._cons_parse(returns, asset_weight_cons, port_stats_cons)
        asset_list = returns.columns.tolist() 
        if targets == []:
            targets = returns.shape[1] * [1/returns.shape[1]]
        V = np.asmatrix(returns.corr())
        R = np.asmatrix(returns.mean() * self.trading_day_per_year).T
        x_0 = returns.shape[1] * [1/returns.shape[1]]
        # cons = ({'type': 'eq', 'fun': self._total_weight_constraint},
        # {'type': 'ineq', 'fun': self._long_only_constraint})
        res= sco.minimize(fun=self._calculate_risk_budget_objective_corr_method, x0=x_0, args=[V,targets], method='SLSQP',constraints=cons, options={'disp': True})   
        self.weights = self._get_opt_weight(res, asset_list)
        w = np.asmatrix(res.x)
        print(self.weights)
        rc = self._calculate_risk_budget_var_contribution_corr_method(w, V).tolist()
        rc = [i[0] for i in rc]
        print(rc)
        self.port_stats = self._get_port_stats(res, returns)
   
def get_mean_variance_model_result(opt_method:str='',# 优化目标
                                   annual_vol_limit:float=None,# 预期年化波动率条件
                                   annual_ret_limit:float=None,# 预期年化收益率条件
                                   fund_nav:pd.DataFrame=pd.DataFrame(),# 基金净值
                                   risk_free_rate:float=0.025, # 无风险收益
                                   fund_weight_low_limit_list:list=[], # 权重下限列表
                                   fund_weight_up_limit_list:list=[], # 权重上限列表
                                ):
    try:
        returns = fund_nav.pct_change(1).iloc[1:]
        asset_weight_cons = []
        port_stats_cons = []
        if annual_vol_limit is not None:
            port_stats_cons.append(['std', '<=', annual_vol_limit])
        if annual_ret_limit is not None:
            port_stats_cons.append(['ret', '>=', annual_ret_limit])
        asset_list = returns.columns.tolist()
        for idx, fund_id in enumerate(asset_list):
            _low = fund_weight_low_limit_list[idx]
            _ups = fund_weight_up_limit_list[idx]
            if _low != 0:
                asset_weight_cons.append([fund_id,'>=',_low])
            if _ups != 1:
                asset_weight_cons.append([fund_id,'<=',_ups])
        self = MeanVarianceModel(risk_free=risk_free_rate)
        self.process(returns=returns,
                     opt_method=opt_method,
                     asset_weight_cons=asset_weight_cons,
                     port_stats_cons=port_stats_cons,
                    )

        result = {
                'weights':self.weights.T.to_dict()[0],
                'port_stats': self.port_stats.T.to_dict()[0],
            }
    except Exception as e:
        result = {
            'weights':{},
            'port_stats':{},
        }
        print(f'Failed to get data <err_msg> {e} from asset_calculator.get_mean_variance_model_result')
    return result

def get_black_litterman_model_result(fund_nav:pd.DataFrame=pd.DataFrame(),
                                     market_weights:list=[],
                                     investor_view:list=[],
                                     tau:float=0.0019,
                                     opt_method:str='max_sharpe',
                                     risk_aversion:float=5,
                                     omega_method:str='user_confidences', #'prior_variance'
                                     asset_weight_cons:list=[],
                                     port_stats_cons:list=[],
                                ):
    try:
        returns = fund_nav.pct_change(1).iloc[1:]
        self = BlackLittermanModel()
        self.process(returns,market_weights,investor_view,tau,opt_method,risk_aversion,omega_method,asset_weight_cons,port_stats_cons)
        result = {
                'weights':self.weights.T.to_dict()[0],
                'port_stats': self.port_stats.T.to_dict()[0],
            }

    except Exception as e:
        result = {
            'weights':{},
            'port_stats':{},
        }
        print(f'Failed to get data <err_msg> {e} from asset_calculator.get_black_litterman_model_result')
    return result
                        


def get_risk_parity_model_result(risk_method:str='annual_vol',# mdd var cvar
                                fund_nav:pd.DataFrame=pd.DataFrame(), # 基金净值,
                                risk_free_rate:float=0.025, # 无风险收益
                                ):
    try:
        self = RiskParityModel(risk_free=risk_free_rate)
        self.process(risk_method=risk_method, fund_nav=fund_nav)
        
        result = {
            'weights':self.weights,
            'port_stats': self.port_stats.T.to_dict()[0],
        }
    except Exception as e:
        result = {
            'weights':{},
            'port_stats':{},
        }
        print(f'Failed to get data <err_msg> {e} from asset_calculator.risk_contribution_targets')
    return result


def get_risk_budget_model_result(annual_vol_limit:float=None, # 预期年化波动率条件
                               annual_ret_limit:float=None, # 预期年化收益率条件
                               risk_contribution_targets:list=[], # 风险贡献目标 目标设定1
                               risk_contribution_weights_targets:list=[], # 风险贡献权重目标 目标设定2
                               risk_contribution_low_limit_list:list=[], # 风险贡献占比下限 目标设定3
                               risk_contribution_up_limit_list:list=[], # 风险贡献占比上限 目标设定3
                               annual_vol_list:list=[], # 年化波动率列表
                               fund_nav:pd.DataFrame=pd.DataFrame(), # 基金净值,
                               opt_method:str='', # 优化目标 目标设定3,
                               ):
    try:
        asset_weight_cons = []
        port_stats_cons = []
        if annual_vol_limit is not None:
            port_stats_cons.append(['std', '<=', annual_vol_limit])
        if annual_ret_limit is not None:
            port_stats_cons.append(['ret', '>=', annual_ret_limit])
        if opt_method == '':
            targets = []
            if len(risk_contribution_targets) > 0:
                # 1. 资产风险贡献跟踪误差最小
                # 如果基金风险和大于总共的， 按比例缩小
                if sum(risk_contribution_targets) > annual_vol_limit:
                    print('您设置的各子基金风险贡献目标设定值之和大于预期年化波动率设定值！是否需要做等比例缩小处理')
                    _origin_sum = sum(risk_contribution_targets)
                    risk_contribution_targets = [_ / _origin_sum * annual_vol_limit for _ in risk_contribution_targets]
                else:
                    pass
                _sum = sum(risk_contribution_targets)  
                targets = [_ / _sum for _ in risk_contribution_targets] 
            elif len(risk_contribution_weights_targets) > 0:
                targets = risk_contribution_weights_targets
            self = RiskBudgetModel()
            self.process(fund_nav=fund_nav,
                        targets=targets,
                        asset_weight_cons=asset_weight_cons,
                        port_stats_cons=port_stats_cons)
        else:
            returns = fund_nav.pct_change(1).iloc[1:]
            weight_stats_cons = [annual_vol_list, risk_contribution_low_limit_list, risk_contribution_up_limit_list]
            self = MeanVarianceModel()
            self.process(returns=returns, 
                        opt_method=opt_method,
                        asset_weight_cons=asset_weight_cons,
                        port_stats_cons=port_stats_cons,
                        weight_stats_cons=weight_stats_cons)    
        result = {
            'weights':self.weights.T.to_dict()[0],
            'port_stats': self.port_stats.T.to_dict()[0],
        }
    except Exception as e:
        result = {
            'weights':{},
            'port_stats':{},
        }
        print(f'Failed to get data <err_msg> {e} from asset_calculator.risk_contribution_targets')
    return result


def test_mv():
    # 均值方差模型
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    returns = fund_nav.pct_change(1).iloc[1:]
    # 夏普最大 限制资产权重测试
    opt_method = 'max_sharpe'
    asset_weight_cons = [['000006!0','<=',0.6]]
    port_stats_cons = []
    self = MeanVarianceModel()
    self.process(returns, opt_method, asset_weight_cons,port_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()
    print(w)
    print(p_stats)

    # 收益最大 限制波动率测试
    opt_method = 'max_ret'
    asset_weight_cons = []
    port_stats_cons = [['std', '<=', 0.2]]
    self = MeanVarianceModel()
    self.process(returns,opt_method,asset_weight_cons,port_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()
    print(w)
    print(p_stats)

    # 波动最小 限制收益测试
    opt_method = 'min_std'
    asset_weight_cons = []
    port_stats_cons = [['ret', '>=', 0.2]]
    self = MeanVarianceModel()
    self.process(returns,opt_method,asset_weight_cons,port_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()
    print(w)
    print(p_stats)

def test_bl():
    # black litterman 模型
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    returns = fund_nav.pct_change(1).iloc[1:]
    # 测试用用户设定的信心水平
    market_weights = [36, 8, 56]
    investor_view = [[['000006!0'],['000011!0'],0.03,0.5]]
    tau = 0.0019
    opt_method = 'max_sharpe'
    risk_aversion = 5
    omega_method='user_confidences'#'prior_variance'
    asset_weight_cons = []
    port_stats_cons = []
    self = BlackLittermanModel()
    self.process(returns,market_weights,investor_view,tau,opt_method,risk_aversion,omega_method,asset_weight_cons,port_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()
    print(w)
    print(p_stats)

    # 测试不用用户设定的信心水平
    market_weights = [36, 8, 56]
    investor_view = [[['000006!0'],['000011!0'],0.03,0.5]]
    tau = 0.0019
    opt_method = 'max_sharpe'
    risk_aversion = 5
    omega_method='user_confidences'#'prior_variance'
    asset_weight_cons = []
    port_stats_cons = []
    self = BlackLittermanModel()
    self.process(returns,market_weights,investor_view,tau,opt_method,risk_aversion,omega_method,asset_weight_cons,port_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()
    print(w)
    print(p_stats)

def test_rp():
    # 风险平价模型
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    risk_method = 'annual_vol'
    self = RiskBudgetModel()
    self.process(risk_method, fund_nav)
    w = self.get_opt_weight()
    asset_risk_contribution = self.get_asset_risk()
    print(w)
    print(asset_risk_contribution)

def test_rb():
    # 风险预算模型
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    risk_method = 'annual_vol'
    targets = [0.5,0.25,0.25]
    self = RiskParityModel()
    self.process(risk_method, fund_nav, targets)
    w = self.get_opt_weight()
    asset_risk_contribution = self.get_asset_risk()
    print(w)
    print(asset_risk_contribution)

def test_rb_adv():
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_info = BasicDataApi().get_fund_info(fund_list=fund_list)
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_desc_dic = fund_info.set_index('fund_id').to_dict()['desc_name']
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    asset_weight_cons = []
    port_stats_cons = []
    #port_stats_cons = [['ret','>=',0.5]]
    targets = [0.25,0.5,0.25]
    self = RiskBudgetModel()
    self.process(fund_nav,targets)

def test_rb_adv_2():
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_info = BasicDataApi().get_fund_info(fund_list=fund_list)
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_desc_dic = fund_info.set_index('fund_id').to_dict()['desc_name']
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    #asset_weight_cons = [['000001!0','<=',0.25],['000006!0','<=',0.5],['000011!0','<=',0.5]]
    asset_weight_cons = []
    port_stats_cons = [['std', '<=', 0.20],['ret','>=',0.3]]
    #port_stats_cons = [['ret','>=',0.5]]
    targets = [0.25,0.5,0.25]
    self = RiskBudgetModel()
    self.process(fund_nav,targets,asset_weight_cons,port_stats_cons)
    p_stats = self.get_ports_stats()
    print(p_stats)

def test_rb_with_mv():
    from ..data.api.basic import BasicDataApi
    fund_list = ['000001!0','000006!0','000011!0']
    fund_info = BasicDataApi().get_fund_info(fund_list=fund_list)
    fund_nav_ori = BasicDataApi().get_fund_nav(fund_list=fund_list)
    fund_desc_dic = fund_info.set_index('fund_id').to_dict()['desc_name']
    fund_nav = fund_nav_ori.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value').dropna()
    returns = fund_nav.pct_change(1).iloc[1:]

    risk_list = [19.86, 21.87, 23.93]
    up_list = [0.7,0.6,0.4]
    # 收益最大 限制资产权重测试
    opt_method = 'max_ret'
    asset_weight_cons = []
    port_stats_cons = [['std', '<=', 0.2]]
    weight_stats_cons = [risk_list, up_list]
    self = MeanVarianceModel()
    self.process(returns, opt_method, asset_weight_cons,port_stats_cons,weight_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()


    # 最大夏普 限制资产权重测试
    opt_method = 'max_sharpe'
    asset_weight_cons = []
    port_stats_cons = [['ret','>=',0.2],['std', '<=', 0.2]]
    weight_stats_cons = [risk_list, up_list]
    self = MeanVarianceModel()
    self.process(returns, opt_method, asset_weight_cons,port_stats_cons,weight_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()


    # 最小风险 限制资产权重测试
    opt_method = 'min_std'
    asset_weight_cons = []
    port_stats_cons = [['ret','>=',0.2]]
    weight_stats_cons = [risk_list, up_list]
    self = MeanVarianceModel()
    self.process(returns, opt_method, asset_weight_cons,port_stats_cons,weight_stats_cons)
    w = self.get_opt_weight()
    p_stats = self.get_ports_stats()



import inspect
import enum

INDEX_PRICE_EXTRA_TRADE_DAYS = 8
INDEX_VAL_EXTRA_TRADE_DAYS = 8
FUND_NAV_EXTRA_TRADE_DAYS = 8
FUTURE_PRICE_EXTRA_TRADE_DAYS = 8
QUARTER_UPDATE_DATE_LIST = ('0331', '0630', '0930', '1231')
SEMI_UPDATE_DATE_LIST = ('0630', '1231')


class ReadableEnum(object):
    _code_name_cache = None

    @classmethod
    def init_cache(cls):
        cls._code_name_cache = ({}, {})
        for name in dir(cls):
            attr = getattr(cls, name)
            if name.startswith('_') or inspect.ismethod(attr):
                continue
            # assert attr not in cls._code_name_cache[0], f'failed to register {name}, {attr} is a registed value'
            # assert name not in cls._code_name_cache[1], f'failed to register {attr}, {name} is a registed enum'
            cls._code_name_cache[0][attr] = name
            cls._code_name_cache[1][name] = attr

    @classmethod
    def read(cls, code):
        if cls._code_name_cache is None:
            cls.init_cache()
        return cls._code_name_cache[0].get(code, code)

    @classmethod
    def parse(cls, code):
        if cls._code_name_cache is None:
            cls.init_cache()
        return cls._code_name_cache[1].get(code, code)


class SourceType(enum.IntEnum):
    qieman = 1  # 且慢
    danjuan = 2  # 蛋卷
    em = 3  # 天天基金
    gf_sec = 4  # 广发基金

    # 以下为手工爬取部分
    bcm = 8192  # 交通银行
    ccb = 8193  # 建设银行
    cgb = 8194  # 广发银行
    citics = 8195  # 中信证券
    citicb = 8196  # 中信银行
    cmbc = 8197  # 民生银行
    gtjas = 8198  # 国泰君安
    guosens = 8199  # 国信证券
    icbc = 8200  # 工商银行
    pinganb = 8201  # 平安银行
    spdb = 8202  # 浦发银行

    @staticmethod
    def get_names_of_type():
        return {
            SourceType.qieman: '且慢',
            SourceType.danjuan: '蛋卷',
            SourceType.em: '天天基金',
            SourceType.gf_sec: '广发基金',
            SourceType.bcm: '交通银行',
            SourceType.ccb: '建设银行',
            SourceType.cgb: '广发银行',
            SourceType.citics: '中信证券',
            SourceType.citicb: '中信银行',
            SourceType.cmbc: '民生银行',
            SourceType.gtjas: '国泰君安',
            SourceType.guosens: '国信证券',
            SourceType.icbc: '工商银行',
            SourceType.pinganb: '平安银行',
            SourceType.spdb: '浦发银行',
        }


# 板块类型
class SectorType(enum.IntEnum):
    industry = 1
    topic = 2
    strategy = 3
    scale = 4


# 指数价格数据来源类型
class IndexPriceSource(enum.IntEnum):
    default = 1
    none = 2
    macroeconomic = 3


# 基金收费模式
class CodeFeeMode(enum.IntEnum):
    none = 0
    a_type = 1  # 前端收费
    b_type = 2  # 后端收费


# 股票因子类型
class StockFactorType(enum.IntEnum):
    BASIC = 1  # 基础
    VALUE = 2  # 价值
    GROWTH = 3  # 成长
    FIN_QUALITY = 4  # 财务质量
    LEVERAGE = 5  # 杠杆
    SCALE = 6  # 规模
    MOMENTUM = 7  # 动量
    VOLATILITY = 8  # 波动率
    LIQUIDITY = 9  # 流动性
    ANALYST = 10  # 分析师
    TECH = 11  # 技术
    ALPHA101 = 12  # WorldQuant alpha101
    SPEC = 13  # 一些特殊因子


# 基金因子类型
class FundFactorType(enum.IntEnum):
    BASIC = 1  # 基础
    DERIVED = 2 # 高级
    SCORE = 3 # 分数


# 行业分类类型
class IndClassType(enum.IntEnum):
    SWI1 = 1  # 申万一级行业指数
    SWI2 = 2  # 申万二级行业指数
    SWI3 = 3  # 申万三级行业指数


# 持仓资产类型
class HoldingAssetType(enum.IntEnum):
    MUTUAL = 1  # 公募基金
    HEDGE = 2  # 私募基金


# 资金状态
class FundStatus(enum.IntEnum):
    IN_TRANSIT = 1  # 在途
    DONE = 2  # 完成
    INSENTIVE_DONE = 3  # 业绩计提扣除份额完成


class ExchangeID(ReadableEnum):
    NOT_AVAILABLE = 0
    SSE = 1
    SZE = 2
    HK = 3
    CFFEX = 4
    DCE = 5
    SHFE = 6
    CZCE = 7

    CRYPTO = 100
    HUOBI = 101
    OKEX = 102
    BINANCE = 103
    BITMEX = 104


class Direction(ReadableEnum):
    NOT_AVAILABLE = 0
    BUY = 1
    SELL = 2


class PosiDirection(ReadableEnum):
    NOT_AVAILABLE = 0
    NET = 1
    LONG = 2
    SHORT = 3


class OrderType(ReadableEnum):
    NOT_AVAILABLE = 0
    PLAIN_ORDER_PREFIX = 1
    BASKET_ORDER_PREFIX = 2
    ALGO_ORDER_PREFIX = 3
    # PLAIN
    PLAIN_ORDER = 10
    LIMIT = 11
    MARKET = 12
    FAK = 13
    FOK = 14
    # BASKET
    BASKET_ORDER = 20
    # ALGO
    ALGO_ORDER = 30
    TWAP = 31


class OffsetFlag(ReadableEnum):
    NOT_AVAILABLE = 0
    OPEN = 1
    CLOSE = 2
    FORCE_CLOSE = 3
    CLOSE_TODAY = 4
    CLOSE_YESTERDAY = 5


class OrderStatus(ReadableEnum):
    NOT_AVAILABLE = 0
    UNKNOWN = 1
    PROPOSED = 10
    RESPONDED = 20
    QUEUEING = 30
    NO_TRADE_QUEUEING = 31
    PART_TRADE_QUEUEING = 32
    PENDING_MAX = 39
    # // if status >= 40, it is not pending
    REJECTED = 40  # // router / gateway / exchange reject
    REJECT_BY_ROUTER = 41
    REJECT_BY_GATEWAY = 42
    REJECT_BY_EXCHANGE = 43
    CANCELLED = 50  # // cancelled, no mater all traded or partly traded
    NO_TRADE_CANCELED = 51
    PART_TRADE_CANCELED = 52
    ALL_TRADED = 60
    # // some other middle status...
    TO_CANCEL = 70


class TradingStyle(ReadableEnum):
    NOT_AVAILABLE = 0
    AGGRESSIVE = 1
    NEUTRAL = 2
    CONSERVATIVE = 3


class BarType(ReadableEnum):
    NOT_AVAILABLE = 0
    MIN_1 = 1
    MIN_3 = 2
    MIN_5 = 3
    MIN_15 = 4
    MIN_30 = 5
    HOUR_1 = 10
    HOUR_2 = 11
    HOUR_4 = 12
    HOUR_6 = 13
    HOUR_12 = 14
    DAY_1 = 20
    WEEK_1 = 30
    MONTH_1 = 40
    YEAR_1 = 50

    _type_secs_cache = None

    @classmethod
    def get_seconds(cls, bar_type):
        if cls._type_secs_cache is None:
            cls._type_secs_cache = {
                cls.MIN_1: 1 * 60,
                cls.MIN_3: 3 * 60,
                cls.MIN_5: 5 * 60,
                cls.MIN_15: 15 * 60,
                cls.MIN_30: 30 * 60,
                cls.HOUR_1: 1 * 60 * 60,
                cls.HOUR_2: 2 * 60 * 60,
                cls.HOUR_4: 4 * 60 * 60,
                cls.HOUR_6: 6 * 60 * 60,
                cls.HOUR_12: 12 * 60 * 60,
                cls.DAY_1: 1 * 24 * 60 * 60,
                cls.WEEK_1: 1 * 7 * 24 * 60 * 60,
                cls.MONTH_1: 1 * 30 * 24 * 60 * 60,  # not accurate
                cls.YEAR_1: 1 * 365 * 24 * 60 * 60,  # not accurate
            }
        return cls._type_secs_cache.get(bar_type, -1)


class AssetType(ReadableEnum):
    NOT_AVAILABLE = 0
    #
    TRADITIONAL_ASSET = 10
    STOCK = 11
    FUTURES = 12
    OPTION = 13
    FUND = 14
    #
    CRYPTO_ASSET = 20
    CRYPTO_SPOT = 21
    CRYPTO_CONTRACT = 22
    CRYPTO_MARGIN = 23
    CRYPTO_PERPETUAL = 24
    CRYPTO_CONTRACT_MARGIN = 25  # 为CRYPTO_COIN_MARGIN_CONTRACT对应的margin
    CRYPTO_COIN_MARGIN_CONTRACT = 26  # 这里特指以币计价的合约,确定了为这类asset需再额外确定对应的margin(体现在posItem里)


class ExecRole(ReadableEnum):
    NOT_AVAILABLE = 0
    MAKER = 1
    TAKER = 2


class MarginMode(ReadableEnum):
    NOT_AVAILABLE = 0
    CROSSED = 1
    FIXED = 2

class LongShort():
    # [long, short] +1 top, -1 bottom, 0 None
    Lt      = [1,   0,  'Longtop']
    Lb      = [-1,  0,  'Longbottom']
    St      = [0,   1,  'Shorttop']
    Sb      = [0,   -1, 'Shortbottom']
    Long    = [1,   0,  'Long']
    Short   = [0,   1,  'Short']
    LtSb    = [1,   -1, 'LongtopShortbottom']
    LbSt    = [-1,  1,  'LongbottomShorttop']

    def find_strategy(self, code):
        self._code_name_cache = ({}, {})
        for name in dir(self):
            attr = getattr(self, name)
            if name.startswith('_') or inspect.ismethod(attr):
                continue
            self._code_name_cache[0][attr[2]] = name
            self._code_name_cache[1][name] = attr

        return self._code_name_cache[1][self._code_name_cache[0].get(code, code)]


class ExchangeStatus(ReadableEnum):
    """交易状态指标"""
    OPEN = 0
    SUSPENDED = 1
    LIMITED = 2
    CLOSE = 3


class FOFTradeStatus(ReadableEnum):
    PURCHASE = 1  # 认购
    SUBSCRIBE = 2  # 申购
    REDEEM = 3  # 赎回
    DIVIDEND_CASH = 4  # 现金分红
    DIVIDEND_VOLUME = 5  # 红利再投
    DEDUCT_REWARD = 6  # 划扣业绩报酬
    DEDUCT_REWARD_AND_DIVIDEND_VOLUME = 7  # 划扣业绩报酬且红利再投
    DEDUCT_REWARD_AND_DIVIDEND_CASH = 8  # 划扣业绩报酬且现金分红
    FORCE_PLUS = 9
    FORCE_MINUS = 10


class FOFStatementStatus(ReadableEnum):
    """FOF资产状态"""
    IN_INVESTOR = 1             # 投资者投入
    OUT_INVESTOR = 2            # 投资者取出
    IN_HEDGE = 3                # 底层资产卖出
    OUT_HEDGE = 4               # 底层资产买入
    IN_HEDGE_DIVIDEND = 5       # 底层资产分红
    OUT_FOF_DIVIDEND = 6        # 产品分红
    IN_INTEREST = 7             # 利息入账
    OUT_MANAGEMENT = 8          # 管理费出账
    OUT_CUSTODIAN = 9           # 托管费出账
    OUT_ADMINISTRATIVE = 10     # 行政管理费出账
    IN_OTHER_IN = 11            # 其他入账
    OUT_OTHER_OUT = 12          # 其他出账
    OUT_OTHER_DEBT = 13         # 其他负债
    IN_OTHER_DEBT_OFFSET = 14   # 其他负债冲抵


class FOFTransitStatus(ReadableEnum):
    """FOF在途资金状态"""
    PENDING = 1
    DONE = 2


class FOFOtherRecordStatus(ReadableEnum):
    """FOF其他记录状态"""
    IN = 1
    OUT = 2


class FOFSaleStatue(ReadableEnum):
    """FOF产品状态"""
    COLLECTION = 1          # 募集期
    OPEN = 2                # 开放期
    LOCK = 3                # 锁定期
    CLEAR = 4               # 已清算


class FOFStrategyType(ReadableEnum):
    """FOF产品 策略类型"""
    S1 = 1                  # 主观多头
    S2 = 2                  # 指数增强
    S3 = 3                  # 市场中性
    S4 = 4                  # 管理期货
    S5 = 5                  # 事件驱动
    S6 = 6                  # 宏观对冲
    S7 = 7                  # 高频套利
    S8 = 8                  # 债券策略
    S10 = 10                # 组合策略
    S11 = 11                # 股票多空
    S12 = 12                # 其他（期权等）

    @staticmethod
    def get_type_of_names():
        return {
            '主观多头': FOFStrategyType.S1,
            '指数增强': FOFStrategyType.S2,
            '市场中性': FOFStrategyType.S3,
            '管理期货': FOFStrategyType.S4,
            '事件驱动': FOFStrategyType.S5,
            '宏观对冲': FOFStrategyType.S6,
            '高频套利': FOFStrategyType.S7,
            '债券策略': FOFStrategyType.S8,
            '组合策略': FOFStrategyType.S10,
            '股票多空': FOFStrategyType.S11,
            '其他（期权等）': FOFStrategyType.S12,
        }


FOFStrategyTypeDic = {
    1:'主观多头',
    2:'指数增强',
    3:'市场中性',
    4:'管理期货',
    5:'事件驱动',
    6:'宏观对冲',
    7:'高频套利',
    8:'债券策略',
    10:'组合策略',
    11:'股票多空',
    12:'其他（期权等)',
}
    
class FOFRiskType(ReadableEnum):
    """FOF产品 风险等级"""
    R1 = 1
    R2 = 2
    R3 = 3
    R4 = 4
    R5 = 5


class DividendCalcMethod(ReadableEnum):
    '''分红计算方式'''

    BY_TOTAL_AMOUNT = 1  # 红利总额
    BY_PER_UNIT = 2  # 每单位红利


class FOFIncentiveFeeMode(ReadableEnum):
    ''' 业绩计提方式 '''

    SCST_HIGH_WL_S = 1  # 单人单笔高水位法（扣减份额）
    SCST_HIGH_WL_N = 2  # 单人单笔高水位法（扣减净值）
    INTE_HIGH_WL = 3  # 整体高水位法
    INTE_HIGH_WL_WITH_SCST_R = 4  # 整体高水位配合单人单笔赎回时计提法


class EMailParserType(ReadableEnum):
    ''' 电子邮件 parser 类型'''

    E_PARSER_NAV = 1  # 净值邮件
    E_PARSER_VALUATION = 2  # 估值表邮件



import datetime
import pandas as pd
import numpy as np
import traceback
from typing import Tuple
from functools import partial
from ....constant import INDEX_VAL_EXTRA_TRADE_DAYS
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...view.derived_models import IndexValuationLongTerm
from .derived_data_helper import DerivedDataHelper
from ....util.calculator import *


class IndexValProcessLongTerm:

    YEAR = 242
    WINDOW = YEAR * 10
    MIN_PERIOD = YEAR * 5
    _WINDOW_MINOR = YEAR * 5
    _MIN_PERIOD_MINOR = YEAR * 3
    SCORE_METHOD = {
        'PB百分位': 'pb_pct',
        'PE百分位': 'pe_pct',
        'PS百分位': 'ps_pct',
    }
    TAA_INDEX_LIST = ['hs300', 'csi500', 'gem', 'sp500rmb']

    def __init__(self, data_helper):
        self._data_helper = data_helper
        self.raw_api = RawDataApi()
        self.basic_api = BasicDataApi()

    def init(self, start_time, end_time, update_partially_list: Tuple[str] = ()):
        self.start_time = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_time = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.index_info = self.basic_api.get_index_info()
        # 过滤掉em_id是null的情况
        self.index_info = self.index_info[self.index_info.em_id.notna()]
        if update_partially_list:
            self.index_info = self.index_info[self.index_info.em_id.isin(update_partially_list)]
        em_id_to_index_id = self.index_info[['index_id', 'em_id']].set_index('em_id').to_dict()['index_id']
        self.index_val = self.raw_api.get_em_index_val(self.start_time, self.end_time, self.index_info.em_id.to_list())
        self.index_val['index_id'] = self.index_val['em_id'].map(lambda x: em_id_to_index_id[x])
        self.index_val = self.index_val.drop(columns=['em_id', '_update_time']).set_index(['index_id', 'datetime'])
        self.index_id_score_method_dict = self.index_info[self.index_info['tag_method'] != 'none'][['index_id', 'tag_method']].set_index('index_id').to_dict()['tag_method']

    def calculate(self):
        pctrank = lambda x: x.rank(pct=True).iloc[-1]
        self.index_val = self.index_val.rename(columns={'dividend_yield': 'dy'})
        index_list = self.index_val.index.levels[0].tolist()
        self.result = []
        for index_id in index_list:
            df_i = self.index_val.loc[index_id].copy()
            df_long_term = df_i[['pe_ttm', 'pb_mrq', 'ps_ttm', 'roe']].tail(self.WINDOW + 1)
            df_i.loc[:, 'pe_pct'] = df_long_term['pe_ttm'].rolling(window=self.WINDOW, min_periods=self.MIN_PERIOD).apply(pctrank, raw=False)
            df_i.loc[:, 'pb_pct'] = df_long_term['pb_mrq'].rolling(window=self.WINDOW, min_periods=self.MIN_PERIOD).apply(pctrank, raw=False)
            df_i.loc[:, 'ps_pct'] = df_long_term['ps_ttm'].rolling(window=self.WINDOW, min_periods=self.MIN_PERIOD).apply(pctrank, raw=False)
            df_i.loc[:, 'roe_pct'] = df_long_term['roe'].rolling(window=self.WINDOW, min_periods=self.MIN_PERIOD).apply(pctrank, raw=False)

            df_minor = df_i[['pe_ttm', 'pb_mrq', 'ps_ttm', 'roe']].tail(self._WINDOW_MINOR + 1)
            df_i.loc[:, 'pe_pct_5_3'] = df_minor['pe_ttm'].rolling(window=self._WINDOW_MINOR, min_periods=self._MIN_PERIOD_MINOR).apply(pctrank, raw=False)
            df_i.loc[:, 'pb_pct_5_3'] = df_minor['pb_mrq'].rolling(window=self._WINDOW_MINOR, min_periods=self._MIN_PERIOD_MINOR).apply(pctrank, raw=False)
            df_i.loc[:, 'ps_pct_5_3'] = df_minor['ps_ttm'].rolling(window=self._WINDOW_MINOR, min_periods=self._MIN_PERIOD_MINOR).apply(pctrank, raw=False)
            df_i.loc[:, 'roe_pct_5_3'] = df_minor['roe'].rolling(window=self._WINDOW_MINOR, min_periods=self._MIN_PERIOD_MINOR).apply(pctrank, raw=False)

            df_i['peg_ttm'] = df_i['pe_ttm'] / (df_i['dy'] / df_i['dy'].shift(self.YEAR) - 1)
            df_i.loc[:, 'index_id'] = index_id
            self.result.append(df_i)
            # print(f'finish {index_list.index(index_id)} total: {len(index_list)}')
        self.result = pd.concat(self.result).reset_index()
        score_list = []
        for r in self.result.itertuples(index=False):
            tag_method = self.index_id_score_method_dict.get(r.index_id, None)
            score_data = self.SCORE_METHOD.get(tag_method, 'pe_pct')
            score_list.append(getattr(r, score_data))

        self.result['datetime'] = pd.to_datetime(self.result['datetime'])
        self.result['val_score'] = score_list
        self.result = self.result.replace(np.inf, float('nan')).replace(-np.inf, float('nan'))

        index_list = self.result.index_id.unique()
        dic = {'pb_mrq': 'pb_pct', 'pe_ttm': 'pe_pct', 'roe': 'roe_pct'}
        l = list(dic.values())
        pct_list = [0.3, 0.7]
        res = []
        for index_id in index_list:
            _df = self.result[self.result['index_id'] == index_id].set_index('datetime')
            if _df.shape[0] < 242 * 5:
                _df.loc[:,l] = None
                res.append(_df.reset_index())
            else:
                for index_val, index_pct in dic.items():
                    _res = []
                    _df[index_val].dropna().rolling(window=10 * 242, min_periods=5 * 242).apply(partial(Calculator.roll_pct, pct_list=pct_list, res=_res), raw=True)
                    _df_i = pd.DataFrame(_res, index=_df.index.tolist()[-len(_res):]).rename(columns={0: f'{index_pct}_30', 1: f'{index_pct}_70'})
                    _df = _df.join(_df_i)
                res.append(_df.reset_index())
        self.result = pd.concat(res, axis=0)

    def filter_est_peg(self, x):
        if x is None:
            return
        if x > 10000 or x < -10000:
            return
        return x

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            start_date = start_date_dt - datetime.timedelta(days=3750)  # 10年历史保险起见，多取几天 10*365=3650
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')
            self.init(start_date, end_date)
            self.calculate()
            self.result = self.result.dropna(subset=['pe_ttm', 'pb_mrq', 'ps_ttm'], how='all')
            self.result['datetime'] = self.result.datetime.dt.date
            self.result = self.result[self.result['datetime'] >= (pd.to_datetime(start_date_dt).date() - datetime.timedelta(days=INDEX_VAL_EXTRA_TRADE_DAYS))]
            self.result = self.result[self.result['datetime'] <= pd.to_datetime(end_date_dt).date()]
            self.result['est_peg'] = self.result['est_peg'].map(lambda x: self.filter_est_peg(x))

            # 这里将sp500rmb的估值数据直接拷贝给sp500
            sp500rmb_data = self.result.loc[self.result.index_id == 'sp500rmb', :].copy()
            if not sp500rmb_data.empty:
                sp500rmb_data['index_id'] = 'sp500'
                self.result = self.result.append(sp500rmb_data, ignore_index=False)
            for date in self.result.datetime.unique():
                DerivedDataApi().delete_index_valuation(date, self.result[self.result.datetime == date].index_id.to_list())
            self._data_helper._upload_derived(self.result, IndexValuationLongTerm.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('index_val_long_term')
        return failed_tasks


if __name__ == "__main__":
    IndexValProcessLongTerm(DerivedDataHelper()).process('20210314', '20210314')
